#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import luigi

from disdatluigi.pipe import PipeTask
import disdat.api as api
import disdatluigi.api as dlapi
from tests.conftest import TEST_CONTEXT


""" Tests for dynamic dependencies using Disdat-Luigi

Run tests with 
A.) 1 or more yields
B.) yield of lists
C.) 1 or more workers 
D.) Mixed with other dependencies
E.) Mixed with other upstream yield dependencies 

Need to check lineage as well


"""


class Static(PipeTask):
    n = luigi.IntParameter()
    dyn_yield = luigi.BoolParameter(default=False)

    def pipe_requires(self):
        self.set_bundle_name("Static[{}]".format(self.n))

    def pipe_run(self):
        if self.dyn_yield:
            result = self.yield_dependency(Yielded, params={'n': self.n, 'source': "Static{}Source".format(self.n)})
            yield result
        return self.n * 2


class Yielded(PipeTask):
    n = luigi.IntParameter()
    source = luigi.Parameter(default="RootSource")
    dyn_yield = luigi.BoolParameter(default=False)

    def pipe_requires(self, pipeline_input=None):
        self.set_bundle_name('Yielded{}{}'.format(self.source, self.n))

    def pipe_run(self, pipeline_input=None):
        if self.dyn_yield:
            result = self.yield_dependency(Yielded, params={'n': self.n, 'source': "Yielded{}Source".format(self.n)})
            yield result
        return self.n


class Root(PipeTask):
    """
    Dynamically yield N tasks
    """
    n = luigi.IntParameter(default=0)
    static_deps = luigi.BoolParameter(default=False)
    static_dynamic_deps = luigi.BoolParameter(default=False)

    def pipe_requires(self):
        self.set_bundle_name('root')
        if self.static_deps:
            self.add_dependency('a', Static, params={'n': 0, 'dyn_yield': self.static_dynamic_deps})
            self.add_dependency('b', Static, params={'n': 1, 'dyn_yield': self.static_dynamic_deps})

    def pipe_run(self, a=None, b=None):
        results = []
        for i in range(self.n):
            results.append(self.yield_dependency(Yielded, params={'n': i}))
            yield results[-1]

        print (sum(d.pipe_output for d in results))
        return sum(d.pipe_output for d in results)


class RootYieldList(Root):
    def pipe_run(self, a=None, b=None):
        results = []
        for i in range(self.n):
            results.append(self.yield_dependency(Yielded, params={'n': i}))
        yield results
        print (sum(d.pipe_output for d in results))
        return sum(d.pipe_output for d in results)


def _yield_pipeline(count, workers, roottask, static_deps=False, static_dynamic=False):
    assert len(api.search(TEST_CONTEXT)) == 0, 'Context should be empty'
    dlapi.apply(TEST_CONTEXT, roottask, output_bundle="test_root", params={'n': count,
                                                                         'static_deps': static_deps,
                                                                         'static_dynamic_deps': static_dynamic},
              workers=workers)
    b = api.get(TEST_CONTEXT, "test_root")
    assert b.data == sum(i for i in range(count))


"""  One-by-one yield, with many yields, with many workers """

def test_simple_yield_one():
    _yield_pipeline(1, 1, Root, False)

def test_simple_yield_one_mp():
    _yield_pipeline(1, 3, Root, False)

def test_simple_yield_many():
    _yield_pipeline(10, 1, Root, False)

def test_simple_yield_many_mp():
    _yield_pipeline(10, 3, Root, False)


"""  One-by-one yield, with many yields, with many workers, with static dependencies """

def test_simple_yield_one_statics():
    _yield_pipeline(1, 1, Root, True)

def test_simple_yield_one_mp_statics():
    _yield_pipeline(1, 3, Root, True)

def test_simple_yield_many_statics():
    _yield_pipeline(10, 1, Root, True)

def test_simple_yield_many_mp_statics():
    _yield_pipeline(10, 3, Root, True)

def test_simple_yield_many_mp_statics_dynamic():
    _yield_pipeline(4, 3, Root, True, True)


"""  One-by-one yield, with many yields, with many workers, with static dependencies """

def test_list_yield_one():
    _yield_pipeline(1, 1, RootYieldList, False)

def test_list_yield_one_mp():
    _yield_pipeline(1, 3, RootYieldList, False)

def test_list_yield_many():
    _yield_pipeline(10, 1, RootYieldList, False)

def test_list_yield_many_mp():
    _yield_pipeline(10, 3, RootYieldList, False)


"""  One-by-one yield, with many yields, with many workers, with static dependencies """

def test_list_yield_one_statics():
    _yield_pipeline(1, 1, RootYieldList, True)

def test_list_yield_one_mp_statics():
    _yield_pipeline(1, 3, RootYieldList, True)

def test_list_yield_many_statics():
    _yield_pipeline(10, 1, RootYieldList, True)

def test_list_yield_many_mp_statics():
    _yield_pipeline(10, 3, RootYieldList, True)

def test_list_yield_many_mp_statics_dynamic():
    _yield_pipeline(4, 3, RootYieldList, True, True)


if __name__ == '__main__':
    api.delete_context(TEST_CONTEXT)
    api.context(TEST_CONTEXT)
    #test_simple_yield_one()
    test_simple_yield_many_mp_statics_dynamic()


    api.delete_context(TEST_CONTEXT)
    api.context(TEST_CONTEXT)
    #test_list_yield_many()
    test_list_yield_many_mp_statics_dynamic()



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Test Incremental Push

Use API to create a bundle with some files
push to remote context

author: Kenneth Yocum
"""
import boto3
import luigi
import moto
import pytest

from disdatluigi.pipe import PipeTask
import disdat.api as api
import disdatluigi.api as dlapi
from tests.conftest import TEST_CONTEXT

TEST_REMOTE = '__test_remote_context__'
TEST_BUCKET = 'test-bucket'
TEST_BUCKET_URL = "s3://{}".format(TEST_BUCKET)


class APush(PipeTask):
    def pipe_requires(self):
        self.set_bundle_name('a')

    def pipe_run(self):

        target = self.create_output_file('a.txt')
        with target.open('w') as output:
            output.write('Hi!')
        return {'file': [target.path]}


class BPush(PipeTask):

    n = luigi.IntParameter()

    def pipe_requires(self):
        self.set_bundle_name('b')
        self.add_dependency('a', APush, params={})

    def pipe_run(self, a):
        target = self.create_output_file('b.txt')

        a_path = a['file'][0]
        with open(a_path) as f:
            print(f.read())

        with target.open('w') as output:
            output.write(str(self.n))
        return {'file': [target.path]}


class CPush(PipeTask):

    n = luigi.IntParameter(default=2)

    def pipe_requires(self):
        self.set_bundle_name('c')
        self.add_dependency('b', BPush, params={'n': self.n})

    def pipe_run(self, b=None):
        # Barf!
        raise Exception


@moto.mock_aws
def test_add_with_treat_as_bundle():
    api.delete_context(TEST_CONTEXT)
    api.context(context_name=TEST_CONTEXT)

    # Setup moto s3 resources
    s3_client = boto3.client('s3')
    s3_resource = boto3.resource('s3')
    s3_resource.create_bucket(Bucket=TEST_BUCKET)

    # Make sure bucket is empty
    objects = s3_client.list_objects(Bucket=TEST_BUCKET)
    assert 'Contents' not in objects, 'Bucket should be empty'

    # Bind remote context
    api.remote(TEST_CONTEXT, TEST_REMOTE, TEST_BUCKET_URL)

    # Try to run the pipeline - should fail
    try:
        # Run test pipeline
        dlapi.apply(TEST_CONTEXT, CPush, incremental_push=True)
    except Exception as e:
        pass

    # Get objects from remote
    objects = s3_client.list_objects(Bucket=TEST_BUCKET)
    keys = [o['Key'] for o in objects['Contents']]
    keys = [key.split('/')[-1] for key in keys]

    # Make sure files exist in S3
    for output_file in ['a.txt', 'b.txt']:
        assert output_file in keys, 'Pipeline should have pushed file'

    api.delete_context(TEST_CONTEXT)


if __name__ == '__main__':
    pytest.main([__file__])



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import luigi

from disdatluigi.pipe import PipeTask
import disdat.api as api
import disdatluigi.api as dlapi
from tests.conftest import TEST_CONTEXT


class A(PipeTask):
    def pipe_requires(self, pipeline_input=None):
        self.set_bundle_name('a')

    def pipe_run(self, pipeline_input=None):
        return 2


class B(PipeTask):

    n = luigi.IntParameter()

    def pipe_requires(self, pipeline_input=None):
        self.set_bundle_name('b')

    def pipe_run(self, pipeline_input=None):
        return 2 * self.n


class C(PipeTask):

    n = luigi.IntParameter(default=2)

    def pipe_requires(self, pipeline_input=None):
        self.set_bundle_name('c')
        self.add_dependency('a', A, params={})
        self.add_dependency('b', B, params={'n': self.n})

    def pipe_run(self, pipeline_input=None, a=None, b=None):
        return a + b


def test_single_task():
    assert len(api.search(TEST_CONTEXT)) == 0, 'Context should be empty'

    dlapi.apply(TEST_CONTEXT, A)
    data = api.get(TEST_CONTEXT, 'a').data

    assert data == 2, 'Data did not match output'
    assert type(data) == int, 'Data is not path'
    assert len(api.search(TEST_CONTEXT)) == 1, 'Only one bundle should be present'

    dlapi.apply(TEST_CONTEXT, A)
    assert len(api.search(TEST_CONTEXT)) == 1, 'Only one bundle should be present'


def test_dependant_tasks():
    assert len(api.search(TEST_CONTEXT)) == 0, 'Context should be empty'

    dlapi.apply(TEST_CONTEXT,  C)
    data = api.get(TEST_CONTEXT, 'c').data

    assert data == 6, 'Data did not match output'
    assert type(data) == int, 'Data is not path'
    assert len(api.search(TEST_CONTEXT)) == 3, 'Three bundles should be present'


def test_task_with_parameter():
    assert len(api.search(TEST_CONTEXT)) == 0, 'Context should be empty'

    dlapi.apply(TEST_CONTEXT,  B, params={'n': 10})
    data = api.get(TEST_CONTEXT, 'b').data

    assert data == 20, 'Data did not match output'
    assert type(data) == int, 'Data is not path'
    assert len(api.search(TEST_CONTEXT)) == 1, 'One bundle should be present'

    dlapi.apply(TEST_CONTEXT,  B, params={'n': 20})
    data = api.get(TEST_CONTEXT, 'b').data

    assert data == 40, 'Data did not match output'
    assert type(data) == int, 'Data is not path'
    assert len(api.search(TEST_CONTEXT)) == 2, 'Two bundles should be present'


def test_child_task_with_parameter():
    assert len(api.search(TEST_CONTEXT)) == 0, 'Context should be empty'

    dlapi.apply(TEST_CONTEXT,  C, params={'n': 10})
    data = api.get(TEST_CONTEXT, 'c').data

    assert data == 22, 'Data did not match output'
    assert type(data) == int, 'Data is not path'
    assert len(api.search(TEST_CONTEXT)) == 3, 'Three bundles should be present'

    dlapi.apply(TEST_CONTEXT,  C, params={'n': 20})
    data = api.get(TEST_CONTEXT, 'c').data

    assert data == 42, 'Data did not match output'
    assert type(data) == int, 'Data is not path'
    assert len(api.search(TEST_CONTEXT)) == 5, 'Five bundles should be present'


if __name__ == '__main__':
    dlapi.apply(TEST_CONTEXT, A, workers=1)



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest

from disdatluigi.pipe import PipeTask
from disdatluigi.common import ApplyError
import disdatluigi.api as api
import luigi

from tests.conftest import TEST_CONTEXT

TEST_NAME    = 'test_bundle'


def test(run_test):
    """ Purpose of this test is to have one task that produces a bundle.
    And another task that requires it.

    1.) Create external dep -- also creates PreMaker_auf_datamaker
    dsdt apply - - test_external_bundle.DataMaker --int_array '[1000,2000,3000]'

    2.) Remove Premaker_auf_datamaker
    dsdt rm PreMaker_auf_datamaker

    3.) Try to run Root -- it should find DataMaker but not re-create it or PreMaker_auf_datamaker

    """
    result = None
    try:
        result = api.apply(TEST_CONTEXT, Root, output_bundle='test_api_exit', params={}, force=True, workers=2)
    except ApplyError as e:
        print ("Got ApplyError exception {} result {} ".format(e, e.result))
        assert(e.result['did_work'])
        assert(not e.result['success'])
    finally:
        print("API apply returned {}".format(result))


class FailBate(PipeTask):
    """
    Generate a small data set of possible basketball scores
    """
    unique = luigi.Parameter()

    def pipe_requires(self):
        self.set_bundle_name("GenData")

    def pipe_run(self):

        if self.unique == 1:
            print("Task about to fail . . . ")
            _ = 100 / 0
        elif self.unique == 0:
            pass

        return


class Root(PipeTask):
    """
    Average scores of an upstream task
    """

    def pipe_requires(self):
        """ Depend on GenData """
        self.add_dependency('task_succeeds', FailBate, {'unique': 0})
        self.add_dependency('task_fails', FailBate, {'unique': 1})
        pass

    def pipe_run(self, **kwargs):
        """ Compute average and return as a dictionary """
        return True


if __name__ == "__main__":
    pytest.main([__file__])
    #import multiprocessing as mp
    #mp.set_start_method('fork')
    #test()



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from setuptools import setup, find_packages

setup(
    use_scm_version={
        'write_to': 'disdatluigi/version.py',
        'write_to_template': '__version__ = "{version}"'
    },
    setup_requires=['setuptools_scm'],

    name='disdat-luigi',
    description='Disdat-Luigi: Data-versioned Luigi Workflows',
    author='Ken Yocum',
    author_email='kyocum@gmail.com',
    url='https://github.com/kyocum/disdat-luigi',

    # Choose your license
    license='Apache License, version 2.0',

    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Operating System :: OS Independent',
        'Natural Language :: English',
    ],

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['tests*']),

    # Include non-python files found in each package in the install, if in your MANIFEST.in
    include_package_data=True,

    # If any package contains other resource files, include them here.
    # We copy config/disdat so 'dsdt init' still runs from an installed
    # .egg.  This package_data only works if you're not using "sdist"
    # Otherwise only MANIFEST.in actually works, and then only if include_package_data=True
    package_data={
        '': ['*.json'],
        'disdatluigi': [
            'config/disdatluigi/*',
            'VERSION',
        ],
        'infrastructure': [
            'Dockerfiles/hyperframe_def/*'
            'dockerizer/Makefile',
            'dockerizer/Dockerfiles/*',
            'dockerizer/kickstart/bin/*',
            'dockerizer/kickstart/etc/*',
        ],
    },

    exclude_package_data={
        'disdatluigi': [
            'dockerizer/kickstart/bin/*.pyc',
        ]
    },

    data_files=[('', ['setup.py'])],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed.  If >=, means it worked with the base version.
    # If <= means higher versions broke something.

    install_requires=[
        'disdat>=1.1.3,<1.2',
        'luigi>=3.0,<3.6',
        'boto3>=1.14.49,<2.0',
        'docker>=7.0.0,<7.2.0',
    ],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e '.[dev, rel]'
    extras_require={
        'dev': [
            'pytest',
            'ipython',
            'mock',
            'pylint',
            'coverage',
            'tox',
            'moto>=5',
            'fastparquet',
            's3fs<=0.4.2' # 0.5.0 breaks with aiobotocore and missing AWS headers
        ],
        'rel': [
            'wheel',
        ]
    },

    entry_points={
        'console_scripts': [
            'dsdt_docker = disdatluigi.entrypoints.docker_ep:main'
        ],
        'distutils.commands': [
            "dsdt_distname = disdatluigi.infrastructure.dockerizer.setup_tools_commands:DistributionName",
        ]
    },
)



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Test Incremental Pull

Use API to create a bundle with some files
push to remote context (test assumes that you have AWS credentials for an account.

author: Kenneth Yocum
"""
import boto3
import luigi
import moto
import pytest

from disdatluigi.pipe import PipeTask
import disdat.api as api
import disdatluigi.api as dlapi
from tests.conftest import TEST_CONTEXT


TEST_REMOTE = '__test_remote_context__'
TEST_BUCKET = 'test-bucket'
TEST_BUCKET_URL = "s3://{}".format(TEST_BUCKET)


class AIP(PipeTask):
    def pipe_requires(self):
        self.set_bundle_name('a')

    def pipe_run(self):

        target = self.create_output_file('a.txt')
        with target.open('w') as output:
            output.write('Hi!')
        return {'file': [target.path]}


class BIP(PipeTask):

    n = luigi.IntParameter()

    def pipe_requires(self):
        self.set_bundle_name('b')
        self.add_dependency('a', AIP, params={})

    def pipe_run(self, a):
        target = self.create_output_file('b.txt')

        a_path = a['file'][0]
        with open(a_path) as f:
            print(f.read())

        with target.open('w') as output:
            output.write(str(self.n))
        return {'file': [target.path]}


class CIP(PipeTask):

    n = luigi.IntParameter(default=2)

    def pipe_requires(self):
        self.set_bundle_name('c')
        self.add_dependency('b', BIP, params={'n': self.n})

    def pipe_run(self, b=None):
        target = self.create_output_file('c.txt')
        with target.open('w') as output:
            output.write(str(self.n + 5))
        return {'file': [target.path]}


@moto.mock_aws
def test_add_with_treat_as_bundle():
    api.delete_context(TEST_CONTEXT)
    api.context(context_name=TEST_CONTEXT)
    
    # Setup moto s3 resources
    s3_client = boto3.client('s3')
    s3_resource = boto3.resource('s3')
    s3_resource.create_bucket(Bucket=TEST_BUCKET)

    # Make sure bucket is empty
    objects = s3_client.list_objects(Bucket=TEST_BUCKET)
    assert 'Contents' not in objects, 'Bucket should be empty'

    # Bind remote context
    api.remote(TEST_CONTEXT, TEST_REMOTE, TEST_BUCKET_URL)

    # Run test pipeline
    dlapi.apply(TEST_CONTEXT, CIP)

    # Push bundles to remote
    for bundle_name in ['a', 'b', 'c']:
        assert api.get(TEST_CONTEXT, bundle_name) is not None, 'Bundle should exist'
        api.commit(TEST_CONTEXT, bundle_name)
        api.push(TEST_CONTEXT, bundle_name)

    # Blow away context and recreate
    api.delete_context(TEST_CONTEXT)
    assert TEST_CONTEXT not in api.ls_contexts()

    api.context(context_name=TEST_CONTEXT)
    api.remote(TEST_CONTEXT, TEST_REMOTE, TEST_BUCKET_URL)

    assert api.search(TEST_CONTEXT) == [], 'Context should be empty'

    # Pull bundles from remote
    api.pull(TEST_CONTEXT)

    # Make sure all bundle meta data comes down but data remains in S3
    for bundle_name in ['a', 'b', 'c']:
        bundle = api.get(TEST_CONTEXT, bundle_name)
        assert bundle is not None, 'Bundle should exist'

        data_path = bundle.data['file'][0]
        assert data_path.startswith('s3://'), 'Data should be in S3'

    # Rerun pipeline
    dlapi.apply(TEST_CONTEXT, BIP, params={'n': 100}, incremental_pull=True)

    # Make sure all bundles exist. Bundles a and b should have local paths
    for bundle_name in ['a', 'b', 'c']:
        bundle = api.get(TEST_CONTEXT, bundle_name)
        assert bundle is not None, 'Bundle should exist'

        data_path = bundle.data['file'][0]
        if bundle_name in ['a', 'b']:
            assert not data_path.startswith('s3://'), 'Data should be local'
        else:
            assert data_path.startswith('s3://'), 'Data should be in S3'

    api.delete_context(TEST_CONTEXT)


if __name__ == '__main__':
    #test_add_with_treat_as_bundle()
    # if testing outside of tox, remember, moto hates MP_CONTEXT_TYPE==forkserver.  Go with fork.
    pytest.main([__file__])



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Configuration
"""

import logging
import os
import sys
import shutil
import subprocess

import luigi
from six.moves import configparser
import six

import disdat.common

from disdatluigi import resource
import disdatluigi.config
from disdatluigi import logger as _logger


SYSTEM_CONFIG_DIR = '~/.config/disdatluigi'
PACKAGE_CONFIG_DIR = 'disdatluigi'
LOGGING_FILE = 'logging.conf'
LUIGI_FILE = 'luigi.cfg'
CFG_FILE = 'disdatluigi.cfg'

LOCAL_EXECUTION = 'LOCAL_EXECUTION'  # Docker endpoint env variable if we're running a container locally


class ApplyError(Exception):
    def __init__(self, message, apply_result):
        super(ApplyError, self).__init__(message)
        self.apply_result = apply_result
    @property
    def result(self):
        return self.apply_result


class ExtDepError(Exception):
    def __init__(self, message):
        super(ExtDepError, self).__init__(message)


def error(msg, *args, **kwargs):
    _logger.error(msg, *args, **kwargs)
    sys.exit(1)


def apply_handle_result(apply_result, raise_not_exit=False):
    """ Execute an appropriate sys.exit() call based on the dictionary
    returned by apply.

    Args:
        apply_result(dict): Has keys 'success' and 'did_work' that give Boolean values.
        raise_not_exit (bool): Raise ApplyException instead of performing sys.exit

    Returns:
        None

    """

    if apply_result['success']:
        if raise_not_exit:
            pass
        else:
            sys.exit(None) # None yields exit value of 0
    else:
        error_str = "Disdat Apply ran, but one or more tasks failed or missing dependencies."
        if raise_not_exit:
            raise ApplyError(error_str, apply_result)
        else:
            sys.exit(error_str)


class DisdatLuigiConfig(object):
    _instance = None

    def __init__(self, config_dir=None):
        """
        Args:
            config_dir (str): Optional directory from which to get disdatluigi.cfg and luigi.cfg.  Default SYSTEM_CONFIG_DIR
        """

        # Find configuration directory
        if config_dir:
            config_dir = config_dir
        else:
            config_dir = os.path.expanduser(SYSTEM_CONFIG_DIR)

        if not os.path.exists(config_dir):
            # Create a config directory and place default disdatluigi.cfg and luigi.cfg there
            DisdatLuigiConfig.init()

        # Extract individual configuration files
        disdat_cfg = os.path.join(config_dir, CFG_FILE)
        luigi_cfg = os.path.join(config_dir, LUIGI_FILE)

        self.logging_config = None
        self.parser = self._read_configuration_file(disdat_cfg, luigi_cfg)

    @staticmethod
    def instance(config_dir=None):
        """
        Singleton getter

        Args:
            config_dir (str): Optional directory from which to get disdatluigi.cfg and luigi.cfg.  Default SYSTEM_CONFIG_DIR
        """
        if DisdatLuigiConfig._instance is None:
            DisdatLuigiConfig._instance = DisdatLuigiConfig(config_dir=config_dir)
        return DisdatLuigiConfig._instance

    @staticmethod
    def _fix_relative_path(config_file, to_fix_path):
        if not os.path.isabs(to_fix_path):
            return os.path.join(os.path.dirname(config_file), to_fix_path)
        return to_fix_path

    def _read_configuration_file(self, disdat_config_file, luigi_config_file):
        """
        Paths in the config might be relative.  If so, add the prefix to them.
        See if there is a disdatluigi.cfg in cwd.  Then configure disdat and (re)configure logging.
        """
        config = configparser.ConfigParser()
        config.read(disdat_config_file)

        # Set up luigi configuration
        luigi.configuration.get_config().read(luigi_config_file)

        # Tell everything to push warnings through the logging infrastructure
        logging.captureWarnings(True)

        # unfortunately that's not enough -- kill all luigi (and disdat) warnings
        import warnings
        warnings.filterwarnings("ignore")

        return config

    @staticmethod
    def init():
        """
        Create a default configuration in the config directory. This makes a
        disdat folder which contains all configuration files.
        """
        directory = os.path.expanduser(SYSTEM_CONFIG_DIR)

        # Make sure disdat has not already been initialized
        if os.path.exists(directory):
            error(
                'Disdat-Luigi already initialized in {}.'.format(directory)
            )

        # Create outer folder if the system does not have it yet
        path = os.path.dirname(directory)
        if not os.path.exists(path):
            os.mkdir(path)

        # Copy over default configurations
        src = resource.filename(disdatluigi.config, PACKAGE_CONFIG_DIR)
        dst = directory
        shutil.copytree(src, dst)

        # Make sure paths are absolute in luigi config
        luigi_dir = os.path.join(directory, LUIGI_FILE)
        config = configparser.ConfigParser()
        config.read(luigi_dir)
        with open(luigi_dir, 'w') as handle:
            config.write(handle)

#
# subprocess wrapper
#


def do_subprocess(cmd, cli):
    """ Standardize error processing

    Args:
        cmd (str): command to execute
        cli (bool): whether called from CLI (True) or API (False)

    Returns:
        (int): 0 if success, >0 if failure

    """
    output = 'No captured output from running CMD [{}]'.format(cmd)
    try:
        if not cli:
            output = subprocess.check_output(cmd)
        else:
            subprocess.check_call(cmd)
    except subprocess.CalledProcessError as cpe:
        if not cli:
            print (output)
            return cpe.returncode
        raise

    return 0


def do_subprocess_with_output(cmd):
    """ Standardize error processing

    Args:
        cmd (str): command to execute

    Returns:
        (str): output of command

    """
    try:
        output = subprocess.check_output(cmd)
        return output
    except subprocess.CalledProcessError as cpe:
        raise


#
# Make Docker images names from pipeline class names
#


def make_project_image_name(setup_file_path):
    """
    Create a container name from the name field in the setup.py file.
    This uses some setuptools magic.  When you install disdat, we install
    an entrypoint.   It becomes available to anyone using setup tools.
    This extracts the information from the setup.py.

    see disdat/infrastructure/dockerizer/setup_tools_commands.py

    Args:
        setup_file_path (str): The FQP to the setup.py file used to dockerize.

    Returns:
        (str): image name string

    """

    python_command = [
        'python',
        setup_file_path,
        '-q',
        'dsdt_distname'
    ]

    retval = do_subprocess_with_output(python_command).strip()

    # If P3, this may be a byte array.   If P2, if not unicode, convert ...
    retval = six.ensure_str(retval)

    return retval


def make_sagemaker_project_image_name(setup_file_path):
    """ Create the string for the image for this pipeline if it uses sagemaker's
    calling convention

    Args:
        setup_file_path (str): The FQP to the setup.py file used to dockerize

    Returns:
        str: The name of the image + '-sagemaker'
    """

    return make_project_image_name(setup_file_path) + "-sagemaker"


def make_project_repository_name(docker_repository_prefix, setup_file_path):
    return '/'.join(([docker_repository_prefix.strip('/')] if docker_repository_prefix is not None else []) + [make_project_image_name(setup_file_path)])


def make_sagemaker_project_repository_name(docker_repository_prefix, setup_file_path):
    return '/'.join(([docker_repository_prefix.strip('/')] if docker_repository_prefix is not None else []) + [make_sagemaker_project_image_name(setup_file_path)])


#
# Make run commands
#


def get_run_command_parameters(pfs):
    remote = pfs.curr_context.remote_ctxt_url
    if remote is not None:
        remote = remote.replace('/{}'.format(disdat.common.DISDAT_CONTEXT_DIR), '')
        local_ctxt = "{}/{}".format(pfs.curr_context.remote_ctxt, pfs.curr_context.local_ctxt)
    else:
        local_ctxt = "{}".format(pfs.curr_context.local_ctxt)
    return remote, local_ctxt


def make_run_command(
        output_bundle,
        output_bundle_uuid,
        pipe_cls,
        remote,
        context,
        input_tags,
        output_tags,
        force,
        force_all,
        no_pull,
        no_push,
        no_push_int,
        workers,
        pipeline_params
):
    """ Create a list of args.  Note that for execution via run, we always set
    --output-bundle, even though it is optional.   The CLI and API will place a '-'
    if the user does not specify it, which means use the default output bundle name.  Here
    we make sure to pass it through.

    Args:
        output_bundle:
        output_bundle_uuid:
        pipe_cls:
        remote:
        context:
        input_tags:
        output_tags:
        force:
        force_all:
        no_pull:
        no_push:
        no_push_int:
        workers:
        pipeline_params:

    Returns:

    """
    args = [
        '--output-bundle-uuid ', output_bundle_uuid,
        '--output-bundle', output_bundle,
        '--branch', context,
        '--workers', str(workers)
    ]
    if remote:
        args.extend(['--remote', remote])
    if no_pull:
        args += ['--no-pull']
    if no_push:
        args += ['--no-push']
    if no_push_int:
        args += ['--no-push-intermediates']
    if force:
        args += ['--force']
    if force_all:
        args += ['--force-all']
    if len(input_tags) > 0:
        for next_tag in input_tags:
            args += ['--input-tag', next_tag]
    if len(output_tags) > 0:
        for next_tag in output_tags:
            args += ['--output-tag', next_tag]

    args += [str(pipe_cls)]  # The one required argument to the entrypoint

    return [x.strip() for x in args + pipeline_params]


def parse_params(cls, params):
    """
    Create a dictionary of str->str arguments to str->python objects deser'd by Luigi Parameters

    Input is the string "--arg value --arg2 value2"

    Convert to dict {'arg':str,'arg2':str2}

    then

    Convert to dict {'arg':luigi.Parameter.value,'arg2':luigi.Parameter.value2}

    Args:
        cls (type[disdat.pipe.PipeTask]):
        params: from argparse

    Returns:
         dict {'arg':value,'arg2':value2}
    """

    params_str_dict = {k.lstrip('--'): v for k, v in zip(params[::2], params[1::2])}

    return convert_str_params(cls, params_str_dict)


def convert_str_params(cls, params_str):
    """
    This is similar to Luigi.Task.from_str_params(cls, params_str)
    But we don't create the class here, and we outer loop through our params (not the classes
    params).  We just want to convert each of the params that are in the class and in this dictionary
    into the deserialized form.

    NOTE:  This is somewhat dangerous and could break if Luigi changes around
    this code.  The alternative is to use Luigi.load_task() but then we have to ensure
    all the input parameters are "strings" and we have to then put special code
    inside of apply to know when to create a class normally, or create it from the CLI.

    Parameters:
        params_str (dict): dict of str->str.  param name -> value .
    """
    kwargs = {}

    cls_params = {n: p for n, p in cls.get_params()}  # get_params() returns [ (name, param), ... ]

    for param_name, param_str in params_str.items():
        if param_name in cls_params:
            param = cls_params[param_name]
            if isinstance(param_str, list):
                kwargs[param_name] = param._parse_list(param_str)
            else:
                kwargs[param_name] = param.parse(param_str)
        else:
            _logger.error("Parameter {} is not defined in class {}.".format(param_name, cls.__name__))
            raise ValueError("Parameter {} is not defined in class {}.".format(param_name, cls.__name__))

    return kwargs


def setup_exists(fqp_setup):
    """ Check if file exists
    """
    if not os.path.exists(fqp_setup):
        print ("No setup.py found at {}.".format(fqp_setup))
        return False
    return True



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Run dockerized version of this pipe

Run differs from apply.  Apply will run the transform as a Python process.
Run executes the most recently built container.   By default it will run it locally via Docker.
Otherwise containers can be submitted to different backends: AWS Batch, AWS Sagemaker.

"""
import argparse
import inspect
import os
import tempfile
import time
import json
from sys import platform

import docker
import six
import boto3 as b3
from enum import Enum

import disdat.fs as fs
import disdat.common
import disdat.utility.aws_s3 as aws
from disdat.common import DisdatConfig
from disdat import logger as _logger

import disdatluigi.common

_MODULE_NAME = inspect.getmodulename(__file__)


ENTRYPOINT_BIN = 'dsdt_docker'


class Backend(Enum):
    Local = 0
    AWSBatch = 1
    LocalSageMaker = 2
    SageMaker = 3

    @staticmethod
    def default():
        return Backend.Local.name

    @staticmethod
    def options():
        return [i.name for i in list(Backend)]


def _run_local(cli, pipeline_setup_file, arglist, backend):
    """
    Run container locally or run sagemaker container locally
    Args:
        cli (bool): Whether we were called from the CLI or API
        pipeline_setup_file (str): The FQ path to the setup.py used to dockerize the pipeline.
        arglist:
        backend:

    Returns:
        output (str): Returns None if there is a failure

    """

    on_macos = False
    if platform == "darwin":
        on_macos = True

    client = docker.from_env()

    environment = {}
    if 'AWS_PROFILE' in os.environ:
        environment['AWS_PROFILE'] = os.environ['AWS_PROFILE']

    environment[disdatluigi.common.LOCAL_EXECUTION] = 'True'

    # Todo: Local runs do not yet set resource limits, but when they do, we'll have to set this
    #environment['DISDAT_CPU_COUNT'] = vcpus

    volumes = {}
    aws_config_dir = os.getenv('AWS_CONFIG_DIR', os.path.join(os.environ['HOME'], '.aws'))
    if aws_config_dir is not None and os.path.exists(aws_config_dir):
        volumes[aws_config_dir] = {'bind': '/root/.aws', 'mode': 'rw'}

    local_disdat_meta_dir = DisdatConfig.instance().get_meta_dir()
    volumes[local_disdat_meta_dir] = {'bind': '/root/.disdat', 'mode': 'rw'}

    try:
        if backend == Backend.LocalSageMaker:
            pipeline_image_name = disdatluigi.common.make_sagemaker_project_image_name(pipeline_setup_file)
            tempdir = tempfile.mkdtemp()
            with open(os.path.join(tempdir, 'hyperparameters.json'), 'w') as of:
                json.dump(_sagemaker_hyperparameters_from_arglist(arglist), of)
                args = ['train']  # rewrite to just 'train'
                # On mac OS, tempdir returns /var, but is actually /private/var
                # Add /private since it that dir is shared (and not /var) with Docker.
                if on_macos:
                    localdir = os.path.join('/private', tempdir[1:])
                else:
                    localdir = tempdir
                volumes[localdir] = {'bind': '/opt/ml/input/config/', 'mode': 'rw'}
                _logger.info("VOLUMES: {}".format(volumes))
        else:
            # Add the actual command to the arglist (for non-sagemaker runs)
            arglist = [ENTRYPOINT_BIN] + arglist
            pipeline_image_name = disdatluigi.common.make_project_image_name(pipeline_setup_file)

        _logger.debug('Running image {} with arguments {}'.format(pipeline_image_name, arglist))

        stdout = client.containers.run(pipeline_image_name, arglist, detach=False,
                                       environment=environment, init=True, stderr=True, volumes=volumes)
        stdout = six.ensure_str(stdout)
        if cli: print(stdout)
        return stdout
    except docker.errors.ContainerError as ce:
        _logger.error("Internal error running image {}".format(pipeline_image_name))
        _logger.error("Error: {}".format(six.ensure_str(ce.stderr)))
        return six.ensure_str(ce)
    except docker.errors.ImageNotFound:
        _logger.error("Unable to find the docker image {}".format(pipeline_image_name))
        return None


def get_fq_docker_repo_name(is_sagemaker, pipeline_setup_file):
    """
    Produce the fully qualified docker repo name.

    Args:
        is_sagemaker (bool): for sagemaker image
        pipeline_setup_file (str): the path to the setup.py file used to dockerize this pipeline

    Returns:
        (str): The fully qualified docker image repository name
    """
    disdat_config = DisdatConfig.instance()

    repository_prefix = None
    if disdat_config.parser.has_option('docker', 'repository_prefix'):
        repository_prefix = disdat_config.parser.get('docker', 'repository_prefix')
    if is_sagemaker:
        repository_name = disdatluigi.common.make_sagemaker_project_repository_name(repository_prefix, pipeline_setup_file)
    else:
        repository_name = disdatluigi.common.make_project_repository_name(repository_prefix, pipeline_setup_file)

    # Figure out the fully-qualified repository name, i.e., the name
    # including the registry.
    registry_name = disdat_config.parser.get('docker', 'registry').strip('/')
    if registry_name == '*ECR*':
        fq_repository_name = aws.ecr_get_fq_repository_name(repository_name)
    else:
        fq_repository_name = '{}/{}'.format(registry_name, repository_name)

    return fq_repository_name


def _run_aws_batch(arglist, fq_repository_name, job_name, pipeline_image_name,
                   aws_session_token_duration, vcpus, memory,
                   no_submit, job_role_arn):
    """
    Run job on AWS Batch.   Sends to queue configured in disdatluigi.cfg.
    This assumes that you have already created a cluster that will run the jobs
    that have been assigned to that queue.

    Args:
        arglist:
        fq_repository_name (str): The fully qualified docker repository name
        job_name:
        pipeline_image_name:
        aws_session_token_duration:
        vcpus:
        memory:
        no_submit (bool): default False
        job_role_arn (str): Can be None

    Returns:

    """

    def check_role_arn(job_dict, jra):
        """ Check to see if the job desc dictionary contains the same job_role_arn (jra)
        """

        if jra is None:
            if 'jobRoleArn' not in job_dict['containerProperties']:
                return True
        else:
            if 'jobRoleArn' in job_dict['containerProperties']:
                if job_dict['containerProperties']['jobRoleArn'] == jra:
                    return True
        return False

    disdat_config = DisdatConfig.instance()

    # Get the parameter values required to kick off an AWS Batch job.
    # Every batch job must:
    # 1. Have a name
    # 2. Have a job definition that declares which ECR-hosted Docker
    #    image to use.
    # 3. Have a queue that feeds jobs into a compute cluster.
    # 4. The command to execute inside the Docker image; the command
    #    args are more-or-less the same as the ones used to execute
    #    locally using 'dsdt run'

    # Create a Job Definition and upload it.
    # We create per-user job definitions so multiple users do not clobber each other.
    # In addition, we never re-use a job definition, since the user may update
    # the vcpu or memory requirements and those are stuck in the job definition

    job_definition_name = aws.batch_get_job_definition_name(pipeline_image_name)

    if disdat_config.parser.has_option(_MODULE_NAME, 'aws_batch_job_definition'):
        job_definition_name = disdat_config.parser.get(_MODULE_NAME, 'aws_batch_job_definition')

    # TODO: Look through all of history to find one that matches?
    # TODO: Delete old jobs here or let user do it?
    job_definition_obj = aws.batch_get_latest_job_definition(job_definition_name)

    if (job_definition_obj is not None and
            job_definition_obj['containerProperties']['image'] == fq_repository_name and
            job_definition_obj['containerProperties']['vcpus'] == vcpus and
            job_definition_obj['containerProperties']['memory'] == memory and
            check_role_arn(job_definition_obj, job_role_arn)):

        job_definition_fqn = aws.batch_extract_job_definition_fqn(job_definition_obj)

        _logger.info("Re-using prior AWS Batch run job definition : {}".format(job_definition_obj))

    else:
        """ Whether None or doesn't match, make a new one """

        job_definition_obj = aws.batch_register_job_definition(job_definition_name, fq_repository_name,
                                          vcpus=vcpus, memory=memory, job_role_arn=job_role_arn)

        job_definition_fqn = aws.batch_get_job_definition(job_definition_name)

        _logger.info("New AWS Batch run job definition {}".format(job_definition_fqn))

    if no_submit:
        # Return the job description object
        return job_definition_obj

    job_queue = disdat_config.parser.get(_MODULE_NAME, 'aws_batch_queue')

    container_overrides = {'command': arglist}

    # Through the magic of boto3_session_cache, the client in our script
    # here can get at AWS profiles and temporary AWS tokens created in
    # part from MFA tokens generated through the user's shells; we don't
    # have to write special code of our own to deal with authenticating
    # with AWS.
    client = b3.client('batch', region_name=aws.profile_get_region())
    # A bigger problem might be that the IAM role executing the job on
    # a batch EC2 instance might not have access to the S3 remote. To
    # get around this, allow the user to create some temporary AWS
    # credentials.

    if aws_session_token_duration > 0 and job_role_arn is None:
        sts_client = b3.client('sts')
        try:
            token = sts_client.get_session_token(DurationSeconds=aws_session_token_duration)
            credentials = token['Credentials']
            container_overrides['environment'] = [
                {'name': 'AWS_ACCESS_KEY_ID', 'value': credentials['AccessKeyId']},
                {'name': 'AWS_SECRET_ACCESS_KEY', 'value': credentials['SecretAccessKey']},
                {'name': 'AWS_SESSION_TOKEN', 'value': credentials['SessionToken']}
            ]
        except Exception as e:
            _logger.debug("Unable to generate an STS token, instead trying users default credentials...")
            credentials = b3.session.Session().get_credentials()
            container_overrides['environment'] = [
                {'name': 'AWS_ACCESS_KEY_ID', 'value': credentials.access_key},
                {'name': 'AWS_SECRET_ACCESS_KEY', 'value': credentials.secret_key},
                {'name': 'AWS_SESSION_TOKEN', 'value': credentials.token}
            ]

    container_overrides['environment'].append({'name': 'DISDAT_CPU_COUNT', 'value': str(vcpus)})

    job = client.submit_job(jobName=job_name, jobDefinition=job_definition_fqn, jobQueue=job_queue,
                            containerOverrides=container_overrides)

    status = job['ResponseMetadata']['HTTPStatusCode']
    if status == 200:
        _logger.info('Job {} (ID {}) with definition {} submitted to AWS Batch queue {}'.format(job['jobName'], job['jobId'],
                                                                                         job_definition_fqn, job_queue))
        return job
    else:
        _logger.error('Job submission failed: HTTP Status {}'.format())
        return None


def _sagemaker_hyperparameters_from_arglist(arglist):
    """
    Return a dictionary of str:str that emulates
    what SageMaker will return when using boto3 interface.

    Args:
        arglist (list (str)): List of string arguments to entrypoint.py

    Returns:
        (dict (str:str)): Dictionary of string to string
    """

    return {'arglist': json.dumps(arglist)}


def _run_aws_sagemaker(arglist, fq_repository_name, job_name):
    """
    Runs a training job on AWS SageMaker.  This uses the default machine type
    in the disdatluigi.cfg file.

    Args:
        arglist:
        fq_repository_name (str): fully qualified repository name
        job_name:  instance job name

    Returns:
        TrainingJobArn (str)
    """

    disdat_config = DisdatConfig.instance()

    job_name = job_name.replace('_', '-') # b/c SageMaker complains it must be ^[a-zA-Z0-9](-*[a-zA-Z0-9])*

    hyperparameter_dict = _sagemaker_hyperparameters_from_arglist(arglist)

    algorithm_specification = {'TrainingImage': fq_repository_name,
                               'TrainingInputMode': 'File'}

    role_arn = disdat_config.parser.get(_MODULE_NAME, 'aws_sagemaker_role_arn')

    input_channel_config = [
        {
            'ChannelName': 'disdat_sagemaker_input_blackhole',
            'DataSource': {
                'S3DataSource': {
                    'S3DataType': 'S3Prefix',
                    'S3Uri': disdat_config.parser.get(_MODULE_NAME, 'aws_sagemaker_s3_input_uri'),
                    'S3DataDistributionType': 'FullyReplicated'
                }
            },
            'ContentType': 'application/javascript',
            'CompressionType': 'None',  # | 'Gzip',
            'RecordWrapperType': 'None' # | 'RecordIO'
        },
    ]

    output_data_config = {'S3OutputPath': os.path.join(
        disdat_config.parser.get(_MODULE_NAME, 'aws_sagemaker_s3_output_uri'), job_name)}

    resource_config = {
            'InstanceType': disdat_config.parser.get(_MODULE_NAME, 'aws_sagemaker_instance_type'),
            'InstanceCount': int(disdat_config.parser.get(_MODULE_NAME, 'aws_sagemaker_instance_count')),
            'VolumeSizeInGB': int(disdat_config.parser.get(_MODULE_NAME, 'aws_sagemaker_volume_sizeGB'))
            # 'VolumeKmsKeyId': 'string'
        }

    vpc_config = None #'SecurityGroupIds': [], 'Subnets': []}

    stopping_condition = {'MaxRuntimeInSeconds': int(disdat_config.parser.get(_MODULE_NAME, 'aws_sagemaker_max_runtime_sec'))}

    tags = [{'Key': 'user', 'Value': 'disdat'},
            {'Key': 'job', 'Value': job_name}]

    if False:
        print("Disdat SageMaker configs")
        print("job name: {}".format(job_name))
        print("hparams: {}".format(hyperparameter_dict))
        print("algorithm: {}".format(algorithm_specification))
        print("Role ARN: {}".format(role_arn))
        print("Input data conf: {}".format(input_channel_config))
        print("Output data conf: {}".format(output_data_config))
        print("Resource conf: {}".format(resource_config))
        print("VPC conf: {}".format(vpc_config))
        print("Stopping condition seconds: {}".format(stopping_condition))
        print("Tags: {}".format(tags))

    client = b3.client('sagemaker', region_name=aws.profile_get_region())

    response = client.create_training_job(
        TrainingJobName=job_name,
        HyperParameters= hyperparameter_dict,
        AlgorithmSpecification=algorithm_specification,
        RoleArn=role_arn,
        InputDataConfig=input_channel_config,
        OutputDataConfig=output_data_config,
        ResourceConfig=resource_config,
        StoppingCondition=stopping_condition,
        Tags=tags
    )

    _logger.info("Disdat SageMaker create_training_job response {}".format(response))
    return response['TrainingJobArn']


def _run(
        output_bundle = '-',
        pipeline_root = '',
        pipeline_args = '',
        pipe_cls = None,
        backend = None,
        input_tags = {},
        output_tags = {},
        force = False,
        force_all = False,
        context = None,
        remote = None,
        no_pull = False,
        no_push = False,
        no_push_int = False,
        vcpus = 1,
        memory = 2000,
        workers = 1,
        no_submit = False,
        job_role_arn = None,
        aws_session_token_duration = 0,
        cli=False):
    """Run the dockerized version of a pipeline.

    Note these are named parameters so we avoid bugs related to argument order.

    Args:
        output_bundle (str): The human name of the output bundle
        pipeline_root (str): The path to the setup.py used to create the container
        pipeline_args: Optional arguments to pass to the pipeline class
        pipe_cls: Name of the pipeline class to run
        backend: The batch execution back-end to use (default
            `Backend.Local`)
        input_tags (list(str)): Find bundle with these tags ['key:value',...]
        output_tags (list(str)): Push result bundle with these tags ['key:value',...]
        force (bool): If `True` force recomputation of all upstream tasks (default `False`)
        force_all (bool): If `True` force recomputation of last task (default `False`)
        context (str): <remote context>/<local context> context string
        remote (str): The remote S3 URL.
        no_pull (bool): Do not pull before executing (start in empty local context)
        no_push (bool): Do not push any new bundles to remote (useful for testing locally)
        no_push_int (bool): Do not push new intermediate bundles to remote
        vcpus (int):  Number of AWS vCPUs the container requests
        memory (int):  Amount of memory container requests in MB
        workers (int): Number of Luigi workers to run tasks in DAG
        no_submit (bool): Produce the AWS job config (for AWS Batch), but do not submit the job
        job_role_arn (str):  The AWS role under which the job should execute
        aws_session_token_duration (int): the number of seconds our temporary credentials should last.
        cli (bool): Whether we called run from the API (buffer output) or the CLI

    Returns:
        job_result (json): A json blob that contains information about the run job.  Error with empty dict.  If backend
        is Sagemaker, return TrainingJobArn.   If backend is AWSBatch, return Batch Job description.   If local, return stdout.
    """
    def assert_or_log(cli, msg):
        if cli:
            _logger.error(msg)
        else:
            assert False, msg

    pfs = fs.DisdatFS()
    pipeline_setup_file = os.path.join(pipeline_root, 'setup.py')

    if not disdatluigi.common.setup_exists(pipeline_setup_file):
        return assert_or_log(cli, "Disdat run: Unable to find setup.py file [{}].".format(pipeline_setup_file))

    # When run in a container, we create the uuid externally to look for a specific result
    output_bundle_uuid = pfs.disdat_uuid()

    # If the user did not specify a context, use the configuration of the current context
    if context is None:
        if not pfs.in_context():
            return assert_or_log(cli, "Disdat run: Not running in a local context. Switch or specify.")
        remote, context = disdatluigi.common.get_run_command_parameters(pfs)

    if remote is None and (not no_push or not no_pull):  # if pulling or pushing, need a remote
        return assert_or_log(cli, "Pushing or pulling bundles with 'run' requires a remote.")

    arglist = disdatluigi.common.make_run_command(output_bundle, output_bundle_uuid, pipe_cls, remote, context,
                                      input_tags, output_tags, force, force_all, no_pull, no_push,
                                      no_push_int, workers, pipeline_args)

    if backend == Backend.AWSBatch or backend == Backend.SageMaker:

        pipeline_image_name = disdatluigi.common.make_project_image_name(pipeline_setup_file)

        job_name = '{}-{}'.format(pipeline_image_name, int(time.time()))

        fq_repository_name = get_fq_docker_repo_name(False, pipeline_setup_file)

        if backend == Backend.AWSBatch:

            # Add the actual command to the arglist
            arglist = [ENTRYPOINT_BIN] + arglist

            retval = _run_aws_batch(arglist,
                                    fq_repository_name,
                                    job_name,
                                    pipeline_image_name,
                                    aws_session_token_duration,
                                    vcpus,
                                    memory,
                                    no_submit,
                                    job_role_arn)
        else:

            fq_repository_name = get_fq_docker_repo_name(True, pipeline_root)

            retval = _run_aws_sagemaker(arglist,
                                        fq_repository_name,
                                        job_name)

    elif backend == Backend.Local or backend == Backend.LocalSageMaker:
        retval = _run_local(cli, pipeline_setup_file, arglist, backend)

    else:
        raise ValueError('Got unrecognized job backend \'{}\': Expected {}'.format(backend, Backend.options()))

    return retval


def add_arg_parser(parsers):
    run_p = parsers.add_parser('run', description="Run a containerized version of transform.")
    run_p.add_argument(
        '--backend',
        default=Backend.default(),
        type=str,
        choices=Backend.options(),
        help='An optional batch execution back-end to use',
    )
    run_p.add_argument(
        "-f",
        "--force",
        action='store_true',
        help="Force recomputation of the last task."
    )
    run_p.add_argument(
        '--force-all',
        action='store_true',
        help="Force re-computation of ALL upstream tasks.")
    run_p.add_argument(
        "--no-submit",
        action='store_true',
        help="For AWS Batch: Do not submit job -- only create the Batch job description."
    )
    run_p.add_argument(
        '--no-push-intermediates',
        action='store_true',
        help='Do not push the intermediate bundles to the remote repository (default is to push)',
        dest='no_push_int'
    )
    run_p.add_argument(
        '--pull',
        action='store_true',
        default=None,
        help="Synchronize local repo and remote before execution. Default False if 'local' backend, else default is True."
    )
    run_p.add_argument(
        '--push',
        action='store_true',
        default=None,
        help="Push bundles to remote context. Default False if 'local' backend, else default is True."
    )
    run_p.add_argument(
        '--use-aws-session-token',
        default=43200, # 12 hours of default time -- for long pipelines!
        type=int,
        help='For AWS Batch: Use temporary AWS session token, valid for AWS_SESSION_TOKEN_DURATION seconds. Default 43200. Set to zero to not use a token.',
        dest='aws_session_token_duration',
    )
    run_p.add_argument('--workers',
                       type=int,
                       default=2,
                       help="The number of Luigi workers to spawn.  Default is 2.")
    run_p.add_argument('--vcpus',
                       type=int,
                       default=2,
                       help="The vCPU count for an AWS Batch container.")
    run_p.add_argument('--memory',
                       type=int,
                       default=4000,
                       help="The memory (MiB) required by this AWS Batch container.")
    run_p.add_argument('--job-role-arn',
                       type=str,
                       default=None,
                       help="For AWS Batch: Use this ARN to indicate the role under which batch containers run.")
    run_p.add_argument('-c', '--context',
                       type=str,
                       default=None,
                       help="'<remote context>/<local context>', else use current context.")
    run_p.add_argument('-r', '--remote',
                       type=str,
                       default=None,
                       help="Remote URL, i.e, 's3://<bucket>/dsdt/', else use remote on current context")
    run_p.add_argument('-it', '--input-tag', nargs=1, type=str, action='append',
                       help="Input bundle tags: '-it authoritative:True -it version:0.7.1'",
                       dest='input_tags')
    run_p.add_argument('-ot', '--output-tag', nargs=1, type=str, action='append',
                       help="Output bundle tags: '-ot authoritative:True -ot version:0.7.1'",
                       dest='output_tags')
    run_p.add_argument('-o', '--output-bundle', type=str, default='-',
                       help="Name output bundle: '-o my.output.bundle'.  Default name is '<TaskName>_<param_hash>'")
    run_p.add_argument("pipeline_root", type=str, help="Root of the Python source tree containing the user-defined transform; must have a setuptools-style setup.py file")
    run_p.add_argument("pipe_cls", type=str, help="User-defined transform, e.g., module.PipeClass")
    run_p.add_argument("pipeline_args", type=str,  nargs=argparse.REMAINDER,
                       help="Optional set of parameters for this pipe '--parameter value'")
    run_p.set_defaults(func=lambda args: run_entry(cli=True, **vars(args)))
    return parsers


def run_entry(cli=False, **kwargs):
    """
    Handles parameter defaults for calling _run()

    From the CLI, the parseargs object has all the arguments
    From the API, the arguments are explicitly set

    Note: pipeline_args is an array of args: ['name',json.dumps(value),'name2',json.dumps(value2)]

    Args:
        kwargs (dict):

    Returns:

    """
    if kwargs['backend'] is not None:
        kwargs['backend'] = Backend[kwargs['backend']]
    else:
        kwargs['backend'] = Backend[Backend.default()]

    # CLI and API only set push or pull.  Translate to no-push no-pull in original run code.
    # If backend == 'Local' then don't pull or push by default
    # TODO: change run() semantics to push pull, not no_push, no_pull.
    if kwargs['push'] is not None:
        kwargs['no_push'] = not kwargs['push']
    else:
        kwargs['no_push'] = True if kwargs['backend'] == Backend.Local else False

    if kwargs['pull'] is not None:
        kwargs['no_pull'] = not kwargs['pull']
    else:
        kwargs['no_pull'] = True if kwargs['backend'] == Backend.Local else False

    # Ensure kwargs only contains the arguments we want when calling _run
    remove_keys = []
    for k in kwargs.keys():
        if k not in _run.__code__.co_varnames:
            remove_keys.append(k)

    for k in remove_keys:
        kwargs.pop(k)

    kwargs['input_tags'] = disdat.common.parse_args_tags(kwargs['input_tags'], to='list')
    kwargs['output_tags'] = disdat.common.parse_args_tags(kwargs['output_tags'], to='list')
    kwargs['cli'] = cli

    return _run(**kwargs)



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
from disdatluigi.pipe import PipeTask
import disdat.api as api
import disdatluigi.api as dlapi
from tests.conftest import TEST_CONTEXT


""" Purpose of this test is to show that if you return nothing, you 
still need to get the input in the downstream task.  See git issue 
 https://github.com/kyocum/disdat/issues/31
 """

WORKERS = 2

class a(PipeTask):
    def pipe_requires(self):
        return

    def pipe_run(self):
        return


class b(PipeTask):
    def pipe_requires(self):
        self.add_dependency('something', a, {})

    def pipe_run(self, something=None):
        print("Return type {}, object: {}".format(type(something), something))
        assert something is None


def test_requires(run_test):
    dlapi.apply(TEST_CONTEXT, b, params={}, workers=WORKERS)


if __name__ == '__main__':
    api.delete_context(TEST_CONTEXT)
    api.context(TEST_CONTEXT)
    test_requires(run_test)
    #pytest.main([__file__])



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
import luigi
from disdatluigi.pipe import PipeTask
import disdat.api as api
import disdatluigi.api as dlapi
from tests.conftest import TEST_CONTEXT

TEST_NAME    = 'test_bundle'


def test(run_test):
    """ This tests if apply force=True and force_all=True re-run everything.
    We have two tasks. One depends on the other.
    force_all should re-run both, force should re-run only the last.
    """

    # first run there should be no bundles
    #assert len(api.search(TEST_CONTEXT)) == 0, 'Context should be empty'
    dlapi.apply(TEST_CONTEXT, A, params={})
    first_B_uuid = api.get(TEST_CONTEXT, 'B').uuid
    first_A_uuid = api.get(TEST_CONTEXT, 'A').uuid

    # second, force re-run last task
    dlapi.apply(TEST_CONTEXT, A, force=True, params={})
    one_B_uuid = api.get(TEST_CONTEXT, 'B').uuid
    one_A_uuid = api.get(TEST_CONTEXT, 'A').uuid
    assert(first_B_uuid == one_B_uuid)
    assert(first_A_uuid != one_A_uuid)

    # second, force all to re-run.
    dlapi.apply(TEST_CONTEXT, A, force_all=True, params={})
    all_B_uuid = api.get(TEST_CONTEXT, 'B').uuid
    all_A_uuid = api.get(TEST_CONTEXT, 'A').uuid
    assert(all_B_uuid != one_B_uuid)
    assert(all_A_uuid != one_A_uuid)

    # third, make sure a force_all doesn't crash if there is an external bundle.
    dlapi.apply(TEST_CONTEXT, A, force_all=True, params={'set_ext_dep': True})
    final_B_uuid = api.get(TEST_CONTEXT, 'B').uuid
    final_A_uuid = api.get(TEST_CONTEXT, 'A').uuid
    assert(final_B_uuid == all_B_uuid)
    assert(final_A_uuid != all_A_uuid)


class B(PipeTask):

    def pipe_requires(self):
        self.set_bundle_name("B")
        return

    def pipe_run(self):
        print ("Task B finished.")

        return True


class A(PipeTask):
    set_ext_dep = luigi.BoolParameter(default=False)

    def pipe_requires(self):
        self.set_bundle_name("A")
        if self.set_ext_dep:
            self.add_external_dependency('B', B, params={})
        else:
            self.add_dependency('B', B, {})

    def pipe_run(self, B=None):
        print ("Task A finished.")
        return


if __name__ == '__main__':
    #test()
    pytest.main([__file__])




#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import uuid
import pytest

from disdat import api
import disdatluigi.api as dlapi
from disdatluigi.pipe import PipeTask
from disdatluigi import common

TAGS = {'tag1': 'omg', 'tag2': 'it works'}


class Source(PipeTask):

    def pipe_requires(self):
        self.set_bundle_name('tagged')

    def pipe_run(self):
        self.add_tags(TAGS)
        return 0


class Destination(PipeTask):

    def pipe_requires(self):
        self.set_bundle_name('output')
        self.add_dependency('tagged', Source, params={})

    def pipe_run(self, tagged):
        tags = self.get_tags('tagged')
        assert tags is not TAGS
        assert tags == TAGS
        return 1


@pytest.fixture
def context():

    try:
        print('ensuring disdat is initialized')
        common.DisdatLuigiConfig.init()
    except:
        print('disdat already initialized, no worries...')

    print('creating temporary local context')
    context = uuid.uuid1().hex
    api.context(context)

    yield context

    print('deleting temporary local context')
    api.delete_context(context)


class TestContext:

    def test_tags(self, context):
        dlapi.apply(context, Destination)


if __name__ == '__main__':
    pytest.main([__file__])



#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import warnings

import disdat.common as common
from disdat.api import _get_context

import disdatluigi.apply
import disdatluigi.run
from disdatluigi.run import Backend, run_entry
from disdatluigi.dockerize import dockerize_entry
from disdatluigi.common import apply_handle_result
from disdatluigi import logger as _logger


def apply(local_context, transform, output_bundle='-',
          input_tags=None, output_tags=None, force=False,
          force_all=False, params=None,
          output_bundle_uuid=None, central_scheduler=False, workers=1,
          incremental_push=False, incremental_pull=False):
    """ Execute a Disdat pipeline natively on the local machine.   Note that `api.run` will execute
    a Disdat pipeline that has been dockerized (either locally or remotely on AWS Batch or AWS Sagemaker)

    Args:
        local_context (str):  The name of the local context in which the pipeline will run in the container
        transform (type[disdat.pipe.PipeTask]): A reference to the Disdat Pipe class
        output_bundle (str):  The name of the output bundle.  Defaults to `<task_name>_<param_hash>`
        input_tags: optional tags dictionary for selecting input bundle
        output_tags: optional tags dictionary to tag output bundle
        force (bool): Force re-running this transform, default False
        force_all (bool): Force re-running ALL transforms, default False
        params: optional parameters dictionary
        output_bundle_uuid: Force UUID of output bundle
        central_scheduler (bool): Use a central scheduler, default False, i.e., use local scheduler
        workers (int): Number of workers, default 1.
        incremental_push (bool): commit and push task bundles as they complete
        incremental_pull (bool): localize bundles from remote as they are required by downstream tasks

    Returns:
        result (int):  0 success, >0 if issue

    """

    # check for deprecated str input for transform
    if isinstance(transform, str):
        msg = ('PipeTask classes should be passed as references, not strings, '
               'support for string inputs will be removed in future versions')
        warnings.warn(msg, DeprecationWarning)
        transform = common.load_class(transform)

    data_context = _get_context(local_context)

    if input_tags is None:
        input_tags = {}

    if output_tags is None:
        output_tags = {}

    if params is None:
        params = {}

    # IF apply raises, let it go up.
    # If API, caller can catch.
    # If CLI, python will exit 1
    result = disdatluigi.apply.apply(output_bundle, params, transform,
                                input_tags, output_tags, force, force_all,
                                output_bundle_uuid=output_bundle_uuid,
                                central_scheduler=central_scheduler,
                                workers=workers,
                                data_context=data_context,
                                incremental_push=incremental_push,
                                incremental_pull=incremental_pull)

    # If no raise, but luigi says not successful
    # If API (here), then raise for caller to catch.
    # For CLI, we exit with 1
    apply_handle_result(result, raise_not_exit=True)

    return result


def run(setup_dir,
        local_context,
        pipe_cls,
        pipeline_args=None,
        output_bundle='-',
        remote_context=None,
        remote_s3_url=None,
        backend=Backend.default(),
        input_tags=None,
        output_tags=None,
        force=False,
        force_all=False,
        pull=None,
        push=None,
        no_push_int=False,
        vcpus=2,
        memory=4000,
        workers=1,
        no_submit=False,
        aws_session_token_duration=42300,
        job_role_arn=None):
    """ Execute a pipeline in a container.  Run locally, on AWS Batch, or AWS Sagemaker

    Simplest execution is with a setup directory (that contains your setup.py), the local context in which to
    execute, and the pipeline to run.   By default this call runs the container locally, reading and writing data only
    to the local context.

    By default this call will assume the remote_context and remote_s3_url of the local context on this system.
    Note that the user must provide both the remote_context and remote_s3_url to override the remote context bound
    to the local context (if any).

    Args:
        setup_dir (str): The directory that contains the setup.py holding the requirements for any pipelines
        local_context (str): The name of the local context in which the pipeline will run in the container
        pipe_cls (str): The pkg.module.class of the root of the pipeline DAG
        pipeline_args (dict): Dictionary of the parameters of the root task
        output_bundle (str): The human name of output bundle
        remote_context (str): The remote context to pull / push bundles during execution. Default is `local_context`
        remote_s3_url (str): The remote's S3 path
        backend : Backend.Local | Backend.AWSBatch. Default Backend.local
        input_tags (dict): str:str dictionary of tags required of the input bundle
        output_tags (dict): str:str dictionary of tags placed on all output bundles (including intermediates)
        force (bool):  Re-run the last pipe task no matter prior outputs
        force_all (bool):  Re-run the entire pipeline no matter prior outputs
        pull (bool): Pull before execution. Default if Backend.Local then False, else True
        push (bool): Push output bundles to remote. Default if Backend.Local then False, else True
        no_push_int (bool):  Do not push intermediate task bundles after execution. Default False
        vcpus (int): Number of virtual CPUs (if backend=`AWSBatch`). Default 2.
        memory (int): Number of MB (if backend='AWSBatch'). Default 2000.
        workers (int):  Number of Luigi workers. Default 1.
        no_submit (bool): If True, just create the AWS Batch Job definition, but do not submit the job
        aws_session_token_duration (int): Seconds lifetime of temporary token (backend='AWSBatch'). Default 42300
        job_role_arn (str): AWS ARN for job execution in a batch container (backend='AWSBatch')

    Returns:
        json (str):

    """

    pipeline_arg_list = []
    if pipeline_args is not None:
        for k,v in pipeline_args.items():
            pipeline_arg_list.append(k)
            pipeline_arg_list.append(json.dumps(v))

    # Set up context as 'remote_name/local_name'
    if remote_context is None:
        assert remote_s3_url is None, "disdat.api.run: user must specify both remote_s3_url and remote_context"
        context = local_context
    else:
        assert remote_s3_url is not None, "disdat.api.run: user must specify both remote_s3_url and remote_context"
        context = "{}/{}".format(remote_context, local_context)

    retval = run_entry(output_bundle=output_bundle,
                       pipeline_root=setup_dir,
                       pipeline_args=pipeline_arg_list,
                       pipe_cls=pipe_cls,
                       backend=backend,
                       input_tags=input_tags,
                       output_tags=output_tags,
                       force=force,
                       force_all=force_all,
                       context=context,
                       remote=remote_s3_url,
                       pull=pull,
                       push=push,
                       no_push_int=no_push_int,
                       vcpus=vcpus,
                       memory=memory,
                       workers=workers,
                       no_submit=no_submit,
                       job_role_arn=job_role_arn,
                       aws_session_token_duration=aws_session_token_duration)

    return retval


def dockerize(setup_dir,
              config_dir=None,
              build=True,
              push=False,
              sagemaker=False):
    """ Create a docker container image using a setup.py and pkg.module.class description of the pipeline.

    Note:
        Users set the os_type and os_version in the disdat.cfg file.
        os_type: The base operating system type for the Docker image
        os_version: The base operating system version for the Docker image

    Args:
        setup_dir (str): The directory that contains the setup.py holding the requirements for any pipelines
        config_dir (str): The directory containing the configuration of .deb packages
        build (bool): If False, just copy files into the Docker build context without building image.
        push (bool):  Push the container to the repository
        sagemaker (bool): Create a Docker image executable as a SageMaker container (instead of a local / AWSBatch container).

    Returns:
        (int): 0 if success, 1 on failure

    """

    retval = dockerize_entry(pipeline_root=setup_dir,
                             config_dir=config_dir,
                             os_type=None,
                             os_version=None,
                             build=build,
                             push=push,
                             sagemaker=sagemaker
                             )

    return retval


def dockerize_get_id(setup_dir):
    """ Retrieve the docker container image identifier

    Args:
        setup_dir (str): The directory that contains the setup.py holding the requirements for any pipelines

    Returns:
        (str): The full docker container image hash
    """
    return dockerize_entry(pipeline_root=setup_dir, get_id=True)
